///  @file rainbow-ring.c
///  @brief The main file of Ring-Rainbow implementation.
///  @source This whole project is a modification of the official reference Rainbow implementation available
///  at the NIST PQC Project, Round 2: https://csrc.nist.gov/CSRC/media/Projects/Post-Quantum-Cryptography/documents/round-2/submissions/Rainbow-Round2.zip
///

#include <stdio.h>
#include <stdint.h>
#include <time.h>
#include "rainbow.h"
#include "rainbow_config.h"
#include "rainbow_blas.h"

#include "utils.h"

#include "rng.h"

#include "api.h"

#define NUMBER_OF_USERS 5
#define NUMBER_OF_EXPERIMENTS 100



int main( int argc , char ** argv )
{
	FILE* output1, *output2;
	char fileName1[80];
	char fileName2[80];
	clock_t start,end;
	sprintf(fileName1, "Rainbow_K=128_NOU=%d_v1=%d_o1=%d_o2=%d_sig",NUMBER_OF_USERS,_V1,_O1,_O2);
	sprintf(fileName2, "Rainbow_K=128_NOU=%d_v1=%d_o1=%d_o2=%d_ver",NUMBER_OF_USERS,_V1,_O1,_O2);
	output1 = fopen(fileName1,"w");
	output2 = fopen(fileName2,"w");

	fprintf(output1, "%s\n", CRYPTO_ALGNAME );
	fprintf(output2, "%s\n", CRYPTO_ALGNAME );
	fprintf(output1, "users: %d\n",NUMBER_OF_USERS);
	fprintf(output2, "users: %d\n",NUMBER_OF_USERS);
        fprintf(output1,"pk size [B]: %lu\n",  NUMBER_OF_USERS*CRYPTO_PUBLICKEYBYTES );
        fprintf(output2,"pk size [B]: %lu\n",  NUMBER_OF_USERS*CRYPTO_PUBLICKEYBYTES );
        fprintf(output1,"signature size [b]: %d\n\n", NUMBER_OF_USERS*_PUB_N_BYTE*8 );
        fprintf(output2,"signature size [b]: %d\n\n", NUMBER_OF_USERS*_PUB_N_BYTE*8 );
	fprintf(output1,"------------------------\n"); fprintf(output2,"------------------------\n"); fflush(output1); fflush(output2);
	int number_of_repeat ;
	unsigned char rnd_seed[48] = {0};
	uint8_t *_sk = (uint8_t*)malloc( (NUMBER_OF_USERS)*CRYPTO_SECRETKEYBYTES);
	uint8_t *_pk = (uint8_t*)malloc( NUMBER_OF_USERS*CRYPTO_PUBLICKEYBYTES );
	FILE * fp;
	clock_t				clock_sign_sum, clock_verification_sum;
	clock_sign_sum = 0;
	clock_verification_sum = 0;
	for ( number_of_repeat = 0;  number_of_repeat < NUMBER_OF_EXPERIMENTS; number_of_repeat++) {
		// set random seed
	
		int rr = byte_from_binfile( rnd_seed , 48 , (4==argc)? argv[3] : "/dev/urandom" );
		if( 0 != rr ) printf("read seed file fail.\n");
		randombytes_init( rnd_seed , NULL , 256 );
	
		/*generation of pairs (publicKey,privateKey) for all users*/
		//start = clock();
		for (int i = 0; i < NUMBER_OF_USERS; i++)
		{
			
			int r = crypto_sign_keypair( _pk + i*(CRYPTO_PUBLICKEYBYTES), _sk + i*(CRYPTO_SECRETKEYBYTES) );

			if( 0 != r ) {
				printf("%s genkey fails.\n", CRYPTO_ALGNAME );
				return -1;
			}
			
		}
		//end = clock();
		//fprintf(stderr,"Keygen: %lf\n",(double)(end-start)/(double)(CLOCKS_PER_SEC)); 
		//printf("generate %s pk/sk success.\n" , CRYPTO_ALGNAME );

		unsigned char cc = 0;
		unsigned char  msg [32];
		unsigned long long mlen = 32;

		int r;
		
		
		/*generate random "document" for signature of size 32B*/
		randombytes(msg, mlen);

		unsigned char digest[_PUB_M_BYTE];
		/*the ring signature will be created by "first" user*/
		/*random vector generation z2, z3, ..., zk*/
		//z_vectors will be the final signature, i.e. z1 = podpis, z2 ... zk will remain the same
		uint8_t *z_vectors = (uint8_t*)malloc( NUMBER_OF_USERS*_PUB_N_BYTE );
		uint8_t *P_z_vectors = (uint8_t*)malloc( NUMBER_OF_USERS*_PUB_M_BYTE );	
		unsigned char w_hat[_PUB_M_BYTE];
		/*Ring Signature Generation - START*/
		start = clock();
		randombytes(z_vectors, NUMBER_OF_USERS*_PUB_N_BYTE);
		
		for (int i = 1; i < NUMBER_OF_USERS; i++)
		{
			batch_quad_trimat_eval( P_z_vectors + i*(_PUB_M_BYTE) , ((pk_t *)(_pk + i*(CRYPTO_PUBLICKEYBYTES)))->pk , z_vectors + i*(_PUB_N_BYTE) , _PUB_N , _PUB_M_BYTE );
		}

		
		hash_msg( digest , _PUB_M_BYTE , msg , mlen );
		//subtract from hash all other Pk(zk)
		for (unsigned i = 1; i < NUMBER_OF_USERS; i++){
			_gf256v_add_u32(digest, P_z_vectors + i*(_PUB_M_BYTE), _PUB_M_BYTE);
	    	}
		//variable digest contains w_hat which we sign with the privateKey of the first user
		rainbow_sign_unsalted(z_vectors, ((sk_t *)(_sk)), digest);
		end=clock();
		fprintf(output1,"%lf\n",(double)(end-start)/(double)(CLOCKS_PER_SEC)); fflush(output1);
		/*variable z_vectors now contains the final signature*/
		clock_sign_sum += end-start;
		/*Ring signature Generation - END*/
	
		/*Ring Signature Verification - START*/
		//again calculate the hash of the document
		memset(digest, 0x00, _PUB_M_BYTE);
		memset(P_z_vectors, 0x00, NUMBER_OF_USERS*_PUB_M_BYTE);
		start = clock();
		hash_msg( digest , _PUB_M_BYTE , msg , mlen );
		//evaluate ALL public keys Pi(zi) for all i = 1 ... k
		for (int i = 0; i < NUMBER_OF_USERS; i++)
		{
			batch_quad_trimat_eval( P_z_vectors + i*(_PUB_M_BYTE) , ((pk_t *)(_pk + i*(CRYPTO_PUBLICKEYBYTES)))->pk , z_vectors + i*(_PUB_N_BYTE) , _PUB_N , _PUB_M_BYTE );
		}

		gf256v_set_zero(w_hat, _PUB_M_BYTE);
		for (unsigned i = 0; i < NUMBER_OF_USERS; i++){
			_gf256v_add_u32(w_hat, P_z_vectors + i*(_PUB_M_BYTE), _PUB_M_BYTE);
	    	}	
	
		for(unsigned i=0;i<_PUB_M_BYTE;i++) {
			cc |= (w_hat[i]^digest[i]);
		}
		if (cc == 0)
		{
			//fprintf(stderr,"Verification successful!\n");
		}
		else
		{
			//fprintf(stderr,"Verification FAILED!\n");	
		}	
		/*Ring signature Verification - END*/
		end=clock();
		fprintf(output2,"%lf\n",(double)(end-start)/(double)(CLOCKS_PER_SEC)); fflush(output2);
		clock_verification_sum += end-start;
	}
	fprintf(output1,"------------------------\n"); fprintf(output2,"------------------------\n");
	fprintf(output1,"Avg time sign: %lf\n", (double)clock_sign_sum/((double)(NUMBER_OF_EXPERIMENTS*CLOCKS_PER_SEC)));
	fprintf(output2,"Avg time verify: %lf\n", (double)clock_verification_sum/((double)(NUMBER_OF_EXPERIMENTS*CLOCKS_PER_SEC)));
	free( _sk );
	free( _pk );
	fclose(output1); fclose(output2);
	return 0;
}

